/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      If
Version:   1.0
ID:        ID_IF
Category:  Command
Date:      May 2001
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2001.  All rights reserved.
Short:     Allows conditional statements to be tested in the GUI.
Keywords:  if, else, condition, conditional, statement, test, compare, comparison

###DESCRIPTION###
<p>The If class provides a way of making decision statements inside the user
interface.  It is exceptionally useful for reacting to user events and is
often used in buttons, menus and other clickable items that may require
a decision to be made.  When used in DML scripts, an If object will only be
created if it is requested as static.  In all other cases the DML interpreter
will execute the if statement locally because the creation of an If object
in such instances is unnecessary.</p>

<p>Support for else statements is also provided if an opposite reaction is
required in case the statement is evaluated as being false.  Here are some
examples involving the typical use of both objects:</p>

<pre>
  &lt;onclick&gt;
    &lt;if static statement="[object.x] = 10"&gt;
      ...statement is true...
    &lt;/if&gt;
  &lt;/onclick&gt;
</pre>

<pre>
  &lt;if static test="=[object.x]+20" condition="<" compare="=[comp.x]*2"&gt;
    ...statement is true...
    &lt;else&gt;
      ...statement is false...
    &lt;/else&gt;
  &lt;/if&gt;
</pre>

<p>You will notice that you can use either the Statement field to set the
condition statement, or you can break it down into the Test, Condition and
Compare fields.  In most cases a Statement setting is adequate, but when
using calculations you must break the string into parts.</p>

<p>If you require the statement to be tested dynamically - i.e. if the Test
and Compare parts need to be constantly re-calculated, use the '$' sign to
prevent the early translation of object references.  Example:</p>

<pre>
  &lt;if static test="$=[object.x]+20$" condition="<" compare="$=[comp.x]*2$"&gt;
</pre>

<p>The If class also allows for the testing of object existence.  This can
be useful if you want to check if an object that you have previously created
still exists in the system.  Use of this feature is demonstrated as follows:</p>

<pre>
  &lt;if exists="[myobject]"/&gt;
</pre>

<p>You may also check for file existence, as demonstrated in this example:</p>

<pre>
  &lt;if fileexists="athene:documents/readme.txt"/&gt;
</pre>

<p>When creating an If statement, make sure that the variable types on
either side match - i.e. a number should be compared to a number, not a
string.  If there is a &quot;type mismatch&quot; then the If object will
treat the statement as being false.</p>

###ACTIONS###
Activate:   Activates an If object.
ClosingTag: If the object is non-static, this action will cause the If object to activate and then self-destruct.

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/files/assign.h>
#include <pandora/modules/strings.h>
#include <pandora/input/joyport.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_IF   1.0
#define VER_ELSE 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_IF, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2001.  All rights reserved.",
   "May 2001",
   "If"
};

static struct KernelBase  *KernelBase = 0;
static OBJECTPTR IfClass = 0;
static OBJECTPTR ElseClass = 0;
static OBJECTPTR StringsMod = 0;
static struct StringsBase *StringsBase = 0;

/*****************************************************************************
** If class definition.
*/

struct If {
   OBJECT_HEADER      /* Class header */
   OBJECTID Exists;
   LONG Condition;    /* */
   LONG Static;       /* Set to TRUE if the If object is static */
   LONG Flags;
   
   PRIVATE_FIELDS
   STRING Test;
   STRING Compare;
   MEMORYID TestMID;
   MEMORYID CompareMID;
   WORD CompareType;
   WORD TestType;
};

static ERROR GET_Compare(struct If *, STRING *);
static ERROR GET_Test(struct If *, STRING *);

static ERROR SET_Exists(struct If *, OBJECTID *);
static ERROR SET_Test(struct If *, STRING);
static ERROR SET_Compare(struct If *, STRING);
static ERROR SET_Statement(struct If *, STRING);

#define COND_NOT_EQUAL     1
#define COND_EQUAL         2
#define COND_LESS_THAN     3
#define COND_LESS_EQUAL    4
#define COND_GREATER_THAN  5
#define COND_GREATER_EQUAL 6

#define TYPE_NUMBER    1
#define TYPE_FLOAT     2
#define TYPE_STRING    3
#define TYPE_REFERENCE 4

#define IFF_TRANSLATE 0x00000001

static STRING TypeTable[] = { NULL, "NUMBER", "FLOAT", "STRING", "REFERENCE" };

static struct FieldDef IfFlags[] = {
   { "TRANSLATE",  IFF_TRANSLATE },
   { NULL, NULL }
};

static struct FieldDef ConditionTable[] = {
   { "<>",  COND_NOT_EQUAL },
   { "!=",  COND_NOT_EQUAL },
   { "NOT", COND_NOT_EQUAL },
   { "=",   COND_EQUAL },
   { "==",  COND_EQUAL },
   { "IS",  COND_EQUAL },
   { "<",   COND_LESS_THAN },
   { "<=",  COND_LESS_EQUAL },
   { ">",   COND_GREATER_THAN },
   { ">=",  COND_GREATER_EQUAL },
   { NULL, NULL }
};

static struct FieldArray IfFields[] = {
   { "Exists",    0, FDF_OBJECTID|FDF_RW,        0,                     NULL, SET_Exists },
   { "Condition", 0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&ConditionTable, NULL, NULL },
   { "Static",    0, FDF_LONG|FDF_INIT,          0,                     NULL, NULL },
   { "Flags",     0, FDF_LONGFLAGS|FDF_RI,       (LONG)&IfFlags,        NULL, NULL },
   /*** Virtual fields ***/
   { "Compare",   0, FDF_STRING|FDF_RW,  0, GET_Compare, SET_Compare },
   { "Statement", 0, FDF_STRING|FDF_W,   0, NULL,        SET_Statement },
   { "Test",      0, FDF_STRING|FDF_RW,  0, GET_Test,    SET_Test },
   END_FIELD
};

static ERROR IF_Activate(struct If *, APTR);
static ERROR IF_ClosingTag(struct If *, APTR);
static ERROR IF_Free(struct If *, APTR);
static ERROR IF_ReleaseObject(struct If *, APTR);
static ERROR IF_Init(struct If *, APTR);
static ERROR IF_NewChild(struct If *, struct acNewChild *);
static ERROR IF_NewObject(struct If *, APTR);

static struct ActionArray IfActions[] = {
   { AC_Activate,      IF_Activate },
   { AC_ClosingTag,    IF_ClosingTag },
   { AC_Free,          IF_Free },
   { AC_ReleaseObject, IF_ReleaseObject },
   { AC_Init,          IF_Init },
   { AC_NewChild,      IF_NewChild },
   { AC_NewObject,     IF_NewObject },
   { NULL, NULL }
};

static LONG GetType(STRING);
static LONG TestStatement(struct If *);

/*****************************************************************************
** Else class definition.
*/

struct Else {
   OBJECT_HEADER
   LONG Static;
};

static struct FieldArray ElseFields[] = {
   { "Static", 0, FDF_LONG|FDF_RI, 0, NULL, NULL },
   END_FIELD
};

static ERROR ELSE_Init(struct Else *, APTR);

static struct ActionArray ElseActions[] = {
   { AC_Init, ELSE_Init },
   { NULL, NULL }
};

/*****************************************************************************
** User Instructions.
*/

static BYTE IfUsage[] = {
"The If command is designed for use in object scripts only - i.e.\n\
it will not work from the Command Line Interface.  Please refer\n\
to the If manual in the Pandora Engine SDK for more information\n\
on this command."
};

static BYTE ElseUsage[] = {
"The Else command is designed for use in object scripts only - i.e.\n\
it will not work from the Command Line Interface.  Please refer\n\
to the If manual in the Pandora Engine SDK for more information\n\
on this command."
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (NewPrivateObject(ID_MODULE, NULL, &StringsMod) IS ERR_Okay) {
      SetField(StringsMod, FID_Name, FT_POINTER, "strings");
      if (Action(AC_Init, StringsMod, NULL) IS ERR_Okay) {
         GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase);
      }
   }

   if (CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&IfClass, NULL,
      FID_BaseClassID|TLONG, ID_IF,
      FID_Version|TFLOAT,    VER_IF,
      FID_Name|TPTR,         "If",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        IfUsage,
      FID_Actions|TPTR,      IfActions,
      FID_Fields|TPTR,       IfFields,
      FID_Size|TLONG,        sizeof(struct If),
      TAGEND) != ERR_Okay) return(ERR_AddClass);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ElseClass, NULL,
      FID_BaseClassID|TLONG, ID_ELSE,
      FID_Version|TFLOAT,    VER_ELSE,
      FID_Name|TPTR,         "Else",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        ElseUsage,
      FID_Actions|TPTR,      ElseActions,
      FID_Fields|TPTR,       ElseFields,
      FID_Size|TLONG,        sizeof(struct Else),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (StringsMod) { Action(AC_Free, StringsMod, NULL); StringsMod = 0; }
   if (IfClass)    { Action(AC_Free, IfClass, NULL);    IfClass    = 0; }
   if (ElseClass)  { Action(AC_Free, ElseClass, NULL);  ElseClass  = 0; }
   return(ERR_Okay);
}

/*****************************************************************************
** If: Activate
*/

static ERROR IF_Activate(struct If *Self, APTR Void)
{
   #define MAX_OBJECTS 50
   OBJECTPTR Child;
   struct ChildEntry list[MAX_OBJECTS], elselist[MAX_OBJECTS];
   LONG i, count, elsecount, j;

   if ((!Self->TestMID) OR (!Self->CompareMID)) return(ObjectError(Self, ERH_Activate, ERR_FieldNotSet));

   /* Test the If statement.  If it is satisfied, activate the children
   ** belonging to the If object.
   */

   if (TestStatement(Self) IS TRUE) {
      count = MAX_OBJECTS;
      if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
         for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
      }
   }
   else {
      count = MAX_OBJECTS;
      if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
         for (i=0; i < count; i++) {
            if (list[i].ClassID IS ID_ELSE) {
               if ((AccessObject(list[i].ObjectID, 0, &Child)) IS ERR_Okay) {
                  elsecount = MAX_OBJECTS;
                  if (ListChildren(list[i].ObjectID, elselist, &elsecount) IS ERR_Okay) {
                     for (j=0; j < elsecount; j++) ActionMsg(AC_Activate, elselist[j].ObjectID, NULL);
                  }
                  ReleaseObject(Child);
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** If: ClosingTag
*/

static ERROR IF_ClosingTag(struct If *If, APTR Void)
{
   if (If->Static IS FALSE) Action(AC_Free, If, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** If: Free
*/

static ERROR IF_Free(struct If *If, APTR Void)
{
   if (If->Test)       { ReleaseMemory(If->Test);      If->Test = NULL; }
   if (If->Compare)    { ReleaseMemory(If->Compare);   If->Compare = NULL; }
   if (If->TestMID)    { FreeMemoryID(If->TestMID);    If->TestMID = NULL; }
   if (If->CompareMID) { FreeMemoryID(If->CompareMID); If->CompareMID = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** If: ReleaseObject
*/

static ERROR IF_ReleaseObject(struct If *If, APTR Void)
{
   if (If->Test)    { ReleaseMemory(If->Test);    If->Test = NULL; }
   if (If->Compare) { ReleaseMemory(If->Compare); If->Compare = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** If: Init
*/

static ERROR IF_Init(struct If *If, APTR Void)
{
   if (!If->TestMID) {
      DPrintF("@If:","The Test field has not been set.");
      return(ERR_FieldNotSet);
   }
   else if (!If->Condition) {
      DPrintF("@If:","The Condition field has not been set.");
      return(ERR_FieldNotSet);
   }
   else if (!If->CompareMID) {
      DPrintF("@If:","The Compare field has not been set.");
      return(ERR_FieldNotSet);
   }

   /* If the If object is not static, perform the test and either fail
   ** or succeed depending on the result.
   */

   if (If->Static IS FALSE) {
      if (TestStatement(If) IS TRUE) return(ERR_Okay);
      else return(ERR_StatementUnsatisfied);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: NewChild()
*/

static ERROR IF_NewChild(struct If *If, struct acNewChild *Args)
{
   OBJECTID OwnerID;
   OBJECTPTR Owner;

   /* If an object is initialised to a non-static If statement, a
   ** pass-through occurs which sees the object being contained
   ** by the If object's container instead.  This feature is intended
   ** for providing correct results when running a script that uses
   ** non-static if statements.
   */

   if (If->Static IS TRUE) return(ERR_Okay);

   OwnerID = GetOwner(If);
   if (AccessObject(OwnerID, 1000, &Owner) IS ERR_Okay) {
      OBJECTPTR NewChild;

      if (AccessObject(Args->NewChildID, 1000, &NewChild) IS ERR_Okay) {
         DPrintF("NewChildAlert()","[If] Child %d will be passed to container %d.", Args->NewChildID, OwnerID);
         SetOwner(NewChild, Owner);
         ReleaseObject(NewChild);
         ReleaseObject(Owner);
         return(ERR_OwnerPassThrough);
      }
      else {
         ReleaseObject(Owner);
         DPrintF("@NewChildAlert:","[If] Could not gain access to new child object %d.", Args->NewChildID);
         return(ERR_ExclusiveDenied);
      }
   }
   else {
      DPrintF("@NewChildAlert:","[If] Could not gain access to my container (%d).", OwnerID);
      return(ERR_ExclusiveDenied);
   }
}

/*****************************************************************************
** If: NewObject
*/

static ERROR IF_NewObject(struct If *If, APTR Args)
{
   If->Static = FALSE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Compare
Short:  This field reflects the value that you want to make a comparison to.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>The value that is being used for comparison must be set in this field.  If
the compare string starts with a calculation '=' or object reference '[' then
the value will be calculated dynamically.</p>

###SEE ALSO###
Field: Test

###END###

*****************************************************************************/

static ERROR GET_Compare(struct If *Self, STRING *Value)
{
   if (Self->Compare) {
      *Value = Self->Compare;
      return(ERR_Okay);
   }
   else if (!Self->CompareMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->CompareMID, MEM_READ, (void **)&Self->Compare) IS ERR_Okay) {
      *Value = Self->Compare;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Compare(struct If *Self, STRING Value)
{
   LONG i;

   if (Self->Compare)    { ReleaseMemory(Self->Compare);   Self->Compare = NULL; }
   if (Self->CompareMID) { FreeMemoryID(Self->CompareMID); Self->CompareMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Compare, &Self->CompareMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Compare[i] = Value[i];
         Self->Compare[i] = 0;
         Self->CompareType = GetType(Self->Compare);
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Condition
Short:  Must be set to the condition that you want to use for statement testing.
Type:   LONG / TABLE
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the condition that will be used when an If object
performs its statement testing.  Valid Condition values are:</p>

<pre>
   <b>Value</b>                  <b>Equivalent Strings</b>
   COND_NOT_EQUAL         !=, <>, NOT
   COND_EQUAL             =, ==, IS
   COND_LESS_THAN         <
   COND_LESS_EQUAL        <=
   COND_GREATER_THAN      >
   COND_GREATER_EQUAL     >=
</pre>

###SEE ALSO###
Field: Compare, Test

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Exists
Short:  Special field that can check for the existence of an object.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>If you need to check for the existence of an object, you can set this
field to point to the object that you want to check for.  The object should
exist at the time of setting this field, thereafter the If object will
check for its existence each time that the statement is re-evaluated.</p>

###END###

*****************************************************************************/

static ERROR SET_Exists(struct If *Self, OBJECTID *ObjectID)
{
   if (!*ObjectID) Self->Exists = -1;
   else Self->Exists = *ObjectID;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Statement
Short:  Writing to this field allows you to set the Compare, Condition and Test fields at once.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>This field provides an easy way to set all three of the required If fields
at once.  Once set, the statement field is broken down into the relevant parts
and then each field is set individually.</p>

<p>Setting this field can fail if the statement is not structured correctly
(the required format is "Test Condition Compare", e.g. "10 > 5").  Do not
use calculations when setting the field - this will only cause confusion, so
set the If fields individually if you need to do this.</p>

###END###

*****************************************************************************/

static ERROR SET_Statement(struct If *Self, STRING Statement)
{
   WORD i, conditionpos;
   BYTE save;

   if (!Statement) return(ERR_Failed);

   /*** Split the statement string into Test, Condition and Compare parts ***/

   for (i=0; (Statement[i]) AND (Statement[i] != '!') AND
        (Statement[i] != '>') AND (Statement[i] != '<') AND
        (Statement[i] != '='); i++);

   if (!Statement[i]) {
      DPrintF("@Set.Statement:","[If] Bad string: %s", Statement);
      return(ERR_BadData);
   }
   conditionpos = i;

   /*** Clip trailing spaces and set the Test field ***/
   while ((i > 0) AND (Statement[i-1] IS ' ')) i--;
   save = Statement[i];
   Statement[i] = 0;
   SET_Test(Self, Statement);
   Statement[i] = save;

   /*** Set the Condition field ***/
   for (i=conditionpos; (Statement[i] IS '!') OR (Statement[i] IS '>') OR (Statement[i] IS '<') OR (Statement[i] IS '='); i++);
   if (!Statement[i]) {
      DPrintF("@Set.Statement:","[If] Bad string: %s", Statement);
      return(ERR_BadData);
   }
   save = Statement[i];
   Statement[i] = 0;
   SetFieldVariable(Self, "Condition", Statement+conditionpos);
   Statement[i] = save;

   /*** Set the Compare field ***/
   while (Statement[i] IS ' ') i++;
   if (Statement[i]) {
      SET_Compare(Self, Statement+i);
   }
   else {
      DPrintF("@Set.Statement:","[If] Bad string: %s", Statement);
      return(ERR_BadData);
   }

   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make the object static.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>By default, an If object will execute itself and then self-destruct
when a closing tag is received.  If you would rather that the object stays
in the system, set this field to TRUE.  If you do this, the only way to get
the If object to test the statement is to call the Activate action.</p>

###NOTE###
<p>The DML processor will only create an If object if the Static field is
specified in the tag-line.  Otherwise, the If statement will be processed
by the DML interpreter in order to give a speed increase.</p>

###END###

****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Test
Short:  This field reflects the value that you want to be tested.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>The value that is to be tested must be set in this field.  If the test
string starts with a calculation '=' or object reference '[' then the value
will be calculated dynamically.</p>

###SEE ALSO###
Field: Compare

###END###

*****************************************************************************/

static ERROR GET_Test(struct If *Self, STRING *Value)
{
   if (Self->Test) {
      *Value = Self->Test;
      return(ERR_Okay);
   }
   else if (!Self->TestMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->TestMID, MEM_READ, (void **)&Self->Test) IS ERR_Okay) {
      *Value = Self->Test;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Test(struct If *Self, STRING Value)
{
   LONG i;

   if (Self->Test)    { ReleaseMemory(Self->Test);   Self->Test = NULL; }
   if (Self->TestMID) { FreeMemoryID(Self->TestMID); Self->TestMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Test, &Self->TestMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Test[i] = Value[i];
         Self->Test[i] = 0;
         Self->TestType = GetType(Self->Test);
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/****************************************************************************/

static LONG GetType(STRING String)
{
   LONG len, i;
   WORD Number, Float;

   len    = StrLength(String);
   Number = TRUE;
   Float  = FALSE;
   for (i=0; (i < len) AND (Number IS TRUE); i++) {
      if (((String[i] < '0') OR (String[i] > '9')) AND (String[i] != '.')) Number = FALSE;
      if (String[i] IS '.') Float = TRUE;
   }

   if ((Float) AND (Number)) return(TYPE_FLOAT);
   else if (Number) return(TYPE_NUMBER);
   else if ((*String IS '[') AND (String[len-1] IS ']')) return(TYPE_REFERENCE);
   else if (*String IS '=') return(TYPE_REFERENCE);
   else if ((*String IS '<') AND (String[len-1] IS '>')) {
      String[0] = '[';
      String[len-1] = ']';
      return(TYPE_REFERENCE);
   }
   else return(TYPE_STRING);
}

/****************************************************************************/

static LONG TestStatement(struct If *Self)
{
   LONG compare_type, test_type, satisfied;
   DOUBLE compare_float, test_float;
   STRING compare_string, test_string;
   BYTE newcompare[1024], newtest[1024], tmp[1024];

   if (!Self->Test)    if (GET_Test(Self, &Self->Test) != ERR_Okay) return(ObjectError(Self, ERH_Activate, ERR_GetField));
   if (!Self->Compare) if (GET_Compare(Self, &Self->Compare) != ERR_Okay) return(ObjectError(Self, ERH_Activate, ERR_GetField));

   /*** Support for "Exists" ***/

   if (Self->Exists) {
      if ((Self->Exists != -1) AND (Self->Exists) AND (CheckObjectExists(Self->Exists, NULL) IS ERR_True)) {
         if (Self->Condition IS COND_NOT_EQUAL) return(FALSE);
         else return(TRUE);
      }
      else {
         if (Self->Condition IS COND_NOT_EQUAL) return(TRUE);
         else return(FALSE);
      }
   }

   /*** Convert the Compare field to its specified type ***/

   satisfied = FALSE;

   if ((Self->Flags & IFF_TRANSLATE) OR (Self->CompareType IS TYPE_REFERENCE)) {
      StrCopy(Self->Compare, newcompare, sizeof(newcompare));
      if (StrTranslate(newcompare, sizeof(newcompare), NULL) != ERR_Okay) return(ERR_Failed);
      if (*newcompare IS '=') {
         if (StrCalculate(newcompare+1, NULL, tmp, sizeof(tmp)) IS ERR_Okay) {
            StrCopy(tmp, newcompare, sizeof(newcompare));
         }
      }
      compare_type = GetType(newcompare);
      compare_string = newcompare;
   }
   else {
      compare_type   = Self->CompareType;
      compare_string = Self->Compare;
   }

   /*** Convert the If->Test to its specified type ***/

   if ((Self->Flags & IFF_TRANSLATE) OR (Self->TestType IS TYPE_REFERENCE)) {
      StrCopy(Self->Test, newtest, sizeof(newtest));
      if (StrTranslate(newtest, sizeof(newtest), NULL) != ERR_Okay) return(ERR_Failed);
      if (*newtest IS '=') {
         if (StrCalculate(newtest+1, NULL, tmp, sizeof(tmp)) IS ERR_Okay) {
            StrCopy(tmp, newtest, sizeof(newtest));
         }
      }
      test_type   = GetType(newtest);
      test_string = newtest;
   }
   else {
      test_type   = Self->TestType;
      test_string = Self->Test;
   }

   /*** Do the comparison ***/

   if (((test_type IS TYPE_FLOAT) OR (test_type IS TYPE_NUMBER)) AND
       ((compare_type IS TYPE_FLOAT) OR (compare_type IS TYPE_NUMBER))) {
      compare_float = StrToFloat(compare_string);
      test_float    = StrToFloat(test_string);
      switch(Self->Condition) {
         case COND_NOT_EQUAL:     if (test_float != compare_float) satisfied = TRUE; break;
         case COND_EQUAL:         if (test_float IS compare_float) satisfied = TRUE; break;
         case COND_LESS_THAN:     if (test_float <  compare_float) satisfied = TRUE; break;
         case COND_LESS_EQUAL:    if (test_float <= compare_float) satisfied = TRUE; break;
         case COND_GREATER_THAN:  if (test_float >  compare_float) satisfied = TRUE; break;
         case COND_GREATER_EQUAL: if (test_float >= compare_float) satisfied = TRUE; break;
         default: DPrintF("@If:","Unsupported condition type %d.", Self->Condition);
      }
   }
   else if ((test_type IS TYPE_STRING) AND (compare_type IS TYPE_STRING)) {
      if (Self->Condition IS COND_EQUAL) {
         if (StrCompare(test_string, compare_string, NULL, STR_MATCHLENGTH) IS ERR_Okay) satisfied = TRUE;
      }
      else if (Self->Condition IS COND_NOT_EQUAL) {
         if (StrCompare(test_string, compare_string, NULL, STR_MATCHLENGTH) != ERR_Okay) satisfied = TRUE;
      }
      else DPrintF("@If:","String comparison for condition %d not possible.", Self->Condition);
   }
   else DPrintF("@If:","Variable type mis-match (%s != %s).", TypeTable[test_type], TypeTable[compare_type]);

   return(satisfied);
}

/****************************************************************************/

#include "else.c"
